import math
from gadapt.ga_model.ga_options import GAOptions
from gadapt.operations.mutation.population_mutation.base_population_mutator import (
    BasePopulationMutator,
)
import gadapt.utils.ga_utils as ga_utils
import statistics as stat


class CostDiversityPopulationMutator(BasePopulationMutator):

    """
    Population mutator based on cost diversity
    """

    def __init__(
        self,
        population_mutator_for_execution: BasePopulationMutator,
    ) -> None:
        super().__init__()
        self._population_mutator_for_execution = population_mutator_for_execution

    def _get_number_of_mutation_cromosomes(
        self, allocated_chromosomes, number_of_mutation_chromosomes
    ) -> int:
        def get_mutation_rate() -> float:
            current_costs = []
            current_min_value = min(allocated_chromosomes, key=lambda x: x.cost_value)
            for c in allocated_chromosomes:
                # current_costs.append(c.cost_value - self.first_cost)
                # current_costs.append(c.cost_value)
                current_costs.append(c.cost_value - current_min_value.cost_value)
            stddev = stat.stdev(current_costs)
            avg = abs(ga_utils.average(current_costs))
            if avg == 0:
                rel_stddev = 0
            else:
                rel_stddev = stddev / avg
            if rel_stddev > 1:
                return 0
            return 1 - rel_stddev

        mutation_rate = get_mutation_rate()
        f_return_value = mutation_rate * float(number_of_mutation_chromosomes)
        return round(f_return_value)

    def _mutate_population(self, population, number_of_mutation_chromosomes):
        if population is None:
            raise Exception("Population must not be null")
        allocated_chromosomes = [c for c in population if not math.isnan(c.cost_value)]
        current_number_of_mutation_chromosomes = (
            self._get_number_of_mutation_cromosomes(
                allocated_chromosomes, number_of_mutation_chromosomes
            )
        )
        return self._population_mutator_for_execution._mutate_population(
            population, current_number_of_mutation_chromosomes
        )



import random
from gadapt.ga_model.gene import Gene
from gadapt.operations.gene_combination.base_gene_combination import BaseGeneCombination


class BlendingGeneCombination(BaseGeneCombination):

    """
    Blending gene combination combines
    gene values from the two parents into new variable values in offsprings.
    One value of the offspring variable comes from a combination of the two
    corresponding values of the parental genes
    """

    def _combine_genes(self, mother_gene: Gene, father_gene: Gene):
        genetic_variable = father_gene.genetic_variable
        val_father = father_gene.variable_value
        val_mother = mother_gene.variable_value
        x = 1
        if val_mother > val_father:
            x = -1
        beta_steps = random.randint(
            0, round(abs((val_father - val_mother) / genetic_variable.step))
        )
        val1 = round(
            val_father - (beta_steps * x) * genetic_variable.step,
            genetic_variable.decimal_places,
        )
        val2 = round(
            val_mother + (beta_steps * x) * genetic_variable.step,
            genetic_variable.decimal_places,
        )
        return val1, val2



from gadapt.factory.ga_factory import GAFactory
from gadapt.adapters.ga_logging.logging_settings import init_logging
from gadapt.ga_model.ga_options import GAOptions
from gadapt.ga_model.ga_results import GAResults
from gadapt.ga_model.population import Population
import gadapt.ga_model.message_levels as message_levels


class GAExecutor:

    """
    Executor for the genetic algorithm

    Args:
        ga_options (GAOptions): Options for GA execution
        factory (GAFactory): Factory for objects creation
    """

    def __init__(self, ga_options: GAOptions, factory: GAFactory) -> None:
        if ga_options is not None:
            self.ga_options = ga_options
        self.factory = factory

    def execute(self) -> GAResults:
        """
        Executes the genetic algorithm
        """
        results = GAResults()
        try:
            init_logging(self.ga_options.logging)
        except Exception as ex:
            results.messages.append(
                (
                    message_levels.WARNING,
                    "Logging failed. Error message: {exc}".format(exc=str(ex)),
                )
            )
            self.ga_options.logging = False
        try:
            chromosome_mutator = self.factory.get_chromosome_mutator()
            population_mutator = self.factory.get_population_mutator()
            exit_checker = self.factory.get_exit_checker()
            cost_finder = self.factory.get_cost_finder()
            population_immigrator = self.factory.get_population_immigrator()
            chromosome_immigrator = self.factory.get_chromosome_immigrator()
            selector = self.factory.get_parent_selector()
            crossover = self.factory.get_crossover()
            variable_updater = variable_updater = self.factory.get_variable_updater()
            population = Population(
                self.ga_options,
                chromosome_mutator=chromosome_mutator,
                population_mutator=population_mutator,
                exit_checker=exit_checker,
                cost_finder=cost_finder,
                population_immigrator=population_immigrator,
                chromosome_immigrator=chromosome_immigrator,
                parent_selector=selector,
                crossover=crossover,
                variable_updater=variable_updater,
            )
            population.find_costs()
            while not population.exit():
                population.immigrate()
                population.mate()
                population.mutate()
                population.find_costs()
            if population.timeout_expired:
                results.messages.append((message_levels.WARNING, "Timeout expired!"))
            best_individual = population.best_individual
            results.min_cost = population.min_cost
            results.number_of_iterations = population.population_generation
            for g in best_individual:
                results.result_values[g.genetic_variable.variable_id] = g.variable_value
        except Exception as ex:
            results.success = False
            results.messages.append((message_levels.ERROR, str(ex)))
        return results



from abc import ABC, abstractmethod


class BaseChromosomeMutator(ABC):

    """
    Base class for the mutation of chromosome.
    Mutates specific genes in the chromosome.
    """

    def mutate(self, c, number_of_mutation_genes: int):
        """
        Mutates genes in the chromosome.
        Args:
            number_of_mutation_genes (int): Number of mutated genes
        """
        self._before_mutated(c)
        self._mutate_chromosome(c, number_of_mutation_genes)
        self._chromosome_mutated(c)

    @abstractmethod
    def _mutate_chromosome(self, c, number_of_mutation_genes: int):
        pass

    def _gene_mutated(self, g, c):
        c.mutated_variables_id_list.append(g.genetic_variable.variable_id)

    def _chromosome_mutated(self, c):
        c.is_mutated = True
        if c.first_mutant_generation == 0:
            c.first_mutant_generation += 1
        c.last_mutant_generation = 1

    def _before_mutated(self, c):
        c.mutated_variables_id_list.clear()



from abc import ABC, abstractmethod


class BaseChromosomeImmigrator(ABC):

    """
    Base class for chromosome immigration
    """

    def immigrate(self, c):
        """
        Makes one chromosome immigrant
        Args:
            c: chromosome to become the immigrant
        """
        self._immigrate_chromosome(c)
        self._chromosome_immigrated(c)

    @abstractmethod
    def _immigrate_chromosome(self, c):
        pass

    def _chromosome_immigrated(self, c):
        c.is_immigrant = True
        if c.first_immigrant_generation == 0:
            c.first_immigrant_generation += 1
        c.last_immigrant_generation = 1
        c.first_mutant_generation = 0
        c.last_mutant_generation = 0
        c.set_chromosome_string_none()



from ast import List
import random
from gadapt.ga_model.chromosome import Chromosome
from gadapt.ga_model.ga_options import GAOptions
from gadapt.operations.mutation.population_mutation.base_population_mutator import (
    BasePopulationMutator,
)
from gadapt.operations.sampling.base_sampling import BaseSampling


class ParentDiversityPopulationMutator(BasePopulationMutator):

    """
    Population mutator based on parent diversity
    """

    def __init__(self, sampling: BaseSampling) -> None:
        super().__init__()
        self._sampling = sampling

    def _sort_key_parent_diversity_random(self, c: Chromosome):
        return (c.parent_diversity, random.random())

    def _mutate_population(self, population, number_of_mutation_chromosomes):
        if population is None:
            raise Exception("Population must not be null")
        unallocated_chromosomes: List[Chromosome] = self._get_unallocated_chromosomes(
            population, self._sort_key_parent_diversity_random
        )
        chromosomes_for_mutation: List[Chromosome] = []
        if population.options.must_mutate_for_same_parents:
            chromosomes_for_mutation = [
                c for c in unallocated_chromosomes if c.parent_diversity == 0
            ]
        chromosomes_for_mutation_count = len(chromosomes_for_mutation)
        rest_number = number_of_mutation_chromosomes - chromosomes_for_mutation_count
        if rest_number > 0:
            if population.options.must_mutate_for_same_parents:
                chromosomes_for_mutation = [
                    c for c in unallocated_chromosomes if (not c.parent_diversity == 0)
                ]
            else:
                chromosomes_for_mutation = [c for c in unallocated_chromosomes]
            chromosomes_for_mutation = self._sampling.get_sample(
                chromosomes_for_mutation, rest_number, lambda c: c.parent_diversity
            )
        for c in chromosomes_for_mutation:
            c.mutate(population.options.number_of_mutation_genes)
        return len(chromosomes_for_mutation)



from gadapt.ga_model.chromosome import Chromosome
from gadapt.ga_model.gene import Gene
import random
from gadapt.operations.mutation.chromosome_mutation.random_chromosome_mutator import (
    RandomChromosomeMutator,
)
from gadapt.operations.sampling.base_sampling import BaseSampling
import gadapt.utils.ga_utils as ga_utils


class CrossDiversityChromosomeMutator(RandomChromosomeMutator):

    """
    Mutation of chromosome based on cross diversity of genetic\
        variables in the population.
    """

    def __init__(self, sampling: BaseSampling) -> None:
        super().__init__()
        self._sampling = sampling

    def _mutate_chromosome(self, c: Chromosome, number_of_mutation_genes: int):
        if number_of_mutation_genes == 0:
            return
        x_genes = [g for g in c]
        x_genes.sort(key=lambda g: -g.genetic_variable.relative_standard_deviation)
        if number_of_mutation_genes > len(x_genes):
            number_of_mutation_genes = len(x_genes)
        number_of_mutation_genes = random.randint(1, number_of_mutation_genes)
        genes_for_mutation = self._sampling.get_sample(
            x_genes,
            number_of_mutation_genes,
            lambda g: g.genetic_variable.relative_standard_deviation,
        )
        for g in genes_for_mutation:
            self._make_mutation(g, c)

    def _make_random_value_below(self, g: Gene):
        if g.variable_value == g.genetic_variable.min_value:
            return g.genetic_variable.make_random_value()
        number_of_steps = random.randint(
            0,
            round(
                (g.variable_value - g.genetic_variable.min_value)
                / g.genetic_variable.step
            ),
        )
        return g.genetic_variable.min_value + number_of_steps * g.genetic_variable.step

    def _make_random_value_above(self, g: Gene):
        if g.variable_value == g.genetic_variable.max_value:
            return g.genetic_variable.make_random_value()
        number_of_steps = random.randint(
            0,
            round(
                (g.genetic_variable.max_value - g.variable_value)
                / g.genetic_variable.step
            ),
        )
        return g.variable_value + number_of_steps * g.genetic_variable.step

    def _make_random_value_below_or_above(self, g: Gene):
        if ga_utils.get_rand_bool():
            return self._make_random_value_above(g)
        return self._make_random_value_below(g)

    def _set_random_value_below_or_above(self, g: Gene, c: Chromosome):
        g.variable_value = round(
            self._make_random_value_below_or_above(g), g.genetic_variable.decimal_places
        )
        self._gene_mutated(g, c)

    def _make_mutation(self, g: Gene, c: Chromosome):
        f = self._get_mutate_func(g)
        f(g, c)

    def _get_mutate_func(self, g: Gene):
        prob = g.genetic_variable.relative_standard_deviation
        if prob > 1.0:
            prob = 1.0
        should_mutate_random = ga_utils.get_rand_bool_with_probability(prob)
        if should_mutate_random:
            return self._set_random_value
        else:
            return self._set_random_value_below_or_above



import logging
import os
from os.path import isfile, join
from gadapt.utils.TimeStampFormatter import TimestampFormatter


def init_logging(is_logging: bool):
    """
    Initializes logging for genetic algorithm
    """

    def get_last_num(s: str) -> int:
        try:
            if not s.startswith("log.log"):
                return -1
            if s == "log.log":
                return 0
            s_last_number = s.rsplit(".", 1)[-1]
            n_last_number = int(s_last_number)
            s.rsplit(".", 1)[-1]
            return n_last_number
        except Exception:
            return -1

    if not is_logging:
        logging.disable(logging.INFO)
        return
    path = os.getcwd() + "\log"
    if not os.path.exists(path):
        os.mkdir(path)
    else:
        onlyfiles = [f for f in os.listdir(path) if isfile(join(path, f))]
        onlyfiles.sort(reverse=True, key=get_last_num)
        for f in onlyfiles:
            if f == "log.log":
                try:
                    os.rename(path + "\\" + f, path + "\\" + "log.log.1")
                except Exception:
                    # print("Unable to rename log file: " + path + "\\" + f)
                    # traceback.print_exc()
                    break
            elif f.startswith("log.log."):
                n_last_number = get_last_num(f)
                if n_last_number == -1:
                    continue
                n_last_number += 1
                try:
                    os.rename(path + "\\" + f, path + "\\log.log." + str(n_last_number))
                except Exception:
                    # print("Unable to rename log file: " + path + "\\" + f)
                    # traceback.print_exc()
                    break
    logpath = path + "\log.log"
    logging.basicConfig(filename=logpath, level=logging.INFO)
    logger = logging.getLogger("")
    handler = logging.FileHandler(logpath)
    handler.setFormatter(
        TimestampFormatter("%(asctime)s - %(levelname)s - %(message)s")
    )
    for h in logger.handlers:
        logger.removeHandler(h)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)



from abc import ABC, abstractmethod
from gadapt.ga_model.chromosome import Chromosome
from gadapt.ga_model.gene import Gene
from gadapt.operations.immigration.chromosome_immigration.base_chromosome_immigrator import (
    BaseChromosomeImmigrator,
)
from gadapt.operations.mutation.chromosome_mutation.base_chromosome_mutator import (
    BaseChromosomeMutator,
)
from gadapt.operations.gene_combination.base_gene_combination import BaseGeneCombination
import gadapt.utils.ga_utils as ga_utils


class BaseCrossover(ABC):

    """Base Crossover Class

    Args:
        gene_combination (BaseGeneCombination): the algorithm
        for how genes are to be combined
        mutator (BaseChromosomeMutator): mutation algorithm to
        be passed to offspring chromosomes
        immigrator (BaseChromosomeImmigrator): immigration algorithm
        to be passed to offspring chromosomes
        mutation_on_both_sides (bool): indicates if offspring chromosomes
        should be mutated on both sides with the same probability
    """

    def __init__(
        self,
        gene_combination: BaseGeneCombination,
        mutator: BaseChromosomeMutator,
        immigrator: BaseChromosomeImmigrator,
        mutation_on_both_sides: bool = True,
    ):
        self._mutation_on_both_sides = True
        self._gene_combination = gene_combination
        self._mutator = mutator
        self._immigrator = immigrator
        self._mutation_on_both_sides = mutation_on_both_sides

    def mate(self, mother: Chromosome, father: Chromosome, population_generation: int):
        """Returns two offspring chromosomes using parents' genetic material

        Args:
            mother (Chromosome): The first chromosome for mating
            father (Chromosome): The second chromosome for mating
            population_generation (int): Current generation in the population

        Returns:
            Chromosome: the first offspring chromosome
            Chromosome: the second offspring chromosome
        """

        def get_genetic_diversity(g_m: Gene, g_f: Gene) -> float:
            return abs(g_m.variable_value - g_f.variable_value) / (
                g_f.genetic_variable.max_value - g_f.genetic_variable.min_value
            )

        if len(mother) != len(father):
            raise Exception("Mother and father must have the same number of genes!")
        offspring1 = Chromosome(self._mutator, self._immigrator, population_generation)
        offspring2 = Chromosome(self._mutator, self._immigrator, population_generation)
        self.number_of_genes = len(father)
        genetic_diversity = []
        for self._current_gene_number in range(self.number_of_genes):
            mother_gene, father_gene = self._get_mother_father_genes(mother, father)
            genetic_variable_father = father_gene.genetic_variable
            genetic_variable_mother = mother_gene.genetic_variable
            if genetic_variable_father != genetic_variable_mother:
                genetic_variable_mother = next(
                    (
                        item.genetic_variable
                        for item in mother
                        if item.genetic_variable == genetic_variable_father
                    ),
                    None,
                )
            if genetic_variable_mother is None:
                raise Exception(
                    "chromosomes in crossover do not have the same structure!"
                )
            genetic_diversity.append(get_genetic_diversity(mother_gene, father_gene))
            var1, var2 = self._combine(mother_gene, father_gene)
            offspring1.add_gene(genetic_variable_father, var1)
            offspring2.add_gene(genetic_variable_father, var2)
        parrents_diversity = round(ga_utils.average(genetic_diversity), 2)
        offspring1.parent_diversity = parrents_diversity
        offspring2.parent_diversity = parrents_diversity
        offspring1.mutation_on_both_sides = self._mutation_on_both_sides
        offspring2.mutation_on_both_sides = self._mutation_on_both_sides
        offspring1.mother_id = mother.chromosome_id
        offspring2.mother_id = mother.chromosome_id
        offspring1.father_id = father.chromosome_id
        offspring2.father_id = father.chromosome_id
        self._increase_generation(offspring1, offspring2, mother, father)
        return offspring1, offspring2

    @abstractmethod
    def _get_mother_father_genes(self, mother: Chromosome, father: Chromosome):
        pass

    @abstractmethod
    def _combine(self, mother_gene: Gene, father_gene: Gene):
        pass

    def _increase_generation(
        self,
        offspring1: Chromosome,
        offspring2: Chromosome,
        mother: Chromosome,
        father: Chromosome,
    ):
        current_generation = mother.chromosome_generation
        if current_generation == 0 or current_generation < father.chromosome_generation:
            current_generation = father.chromosome_generation
        current_generation += 1
        offspring1.chromosome_generation = current_generation
        offspring2.chromosome_generation = current_generation

        current_generation = 0
        if mother.first_mutant_generation > 0 or father.first_mutant_generation > 0:
            current_generation = mother.first_mutant_generation
            if (
                current_generation == 0
                or father.first_mutant_generation > current_generation
            ):
                current_generation = father.first_mutant_generation
            current_generation += 1
        offspring1.first_mutant_generation = current_generation
        offspring2.first_mutant_generation = current_generation

        current_generation = 0
        if mother.last_mutant_generation > 0 or father.last_mutant_generation > 0:
            current_generation = mother.last_mutant_generation
            if current_generation == 0 or (
                father.last_mutant_generation > 0
                and father.last_mutant_generation < current_generation
            ):
                current_generation = father.last_mutant_generation
            current_generation += 1
        offspring1.last_mutant_generation = current_generation
        offspring2.last_mutant_generation = current_generation

        current_generation = 0
        if (
            mother.first_immigrant_generation > 0
            or father.first_immigrant_generation > 0
        ):
            current_generation = mother.first_immigrant_generation
            if (
                current_generation == 0
                or father.first_immigrant_generation > current_generation
            ):
                current_generation = father.first_immigrant_generation
            current_generation += 1
        offspring1.first_immigrant_generation = current_generation
        offspring2.first_immigrant_generation = current_generation

        current_generation = 0
        if mother.last_immigrant_generation > 0 or father.last_immigrant_generation > 0:
            current_generation = mother.last_immigrant_generation
            if current_generation == 0 or (
                father.last_immigrant_generation > 0
                and father.last_immigrant_generation < current_generation
            ):
                current_generation = father.last_immigrant_generation
            current_generation += 1
        offspring1.last_immigrant_generation = current_generation
        offspring2.last_immigrant_generation = current_generation



import random
from typing import List
from gadapt.ga_model.ga_options import GAOptions
from gadapt.operations.mutation.population_mutation.base_population_mutator import (
    BasePopulationMutator,
)


class ComposedPopulationMutator(BasePopulationMutator):
    def __init__(self, options: GAOptions) -> None:
        """
        Population mutator that consists of more different population mutators
        Args:
            options: genetic algorithm options
        """
        super().__init__(options)
        self.mutators: List[BasePopulationMutator] = []

    def append(self, mutator: BasePopulationMutator):
        """
        Appends mutator to the composition of mutators
        """
        self.mutators.append(mutator)

    def _mutate_population(self, population, number_of_mutation_chromosomes):
        if population is None:
            raise Exception("Population must not be null")
        if len(self.mutators) == 0:
            raise Exception("at least one mutator must be added")
        random.shuffle(self.mutators)
        nmc = 0
        for m in self.mutators:
            if nmc < number_of_mutation_chromosomes:
                nmc += m._mutate_population(
                    population, number_of_mutation_chromosomes - nmc
                )



from abc import ABC, abstractmethod


class BaseVariableUpdater(ABC):

    """
    Base class for variable update
    """

    @abstractmethod
    def update_variables(self, population):
        pass



import statistics as stat
from gadapt.ga_model.genetic_variable import GeneticVariable
import gadapt.utils.ga_utils as ga_utils


class CommonVariableUpdater:

    """
    Common variable updater
    """

    def update_variables(self, population):
        def scale_values(gv: GeneticVariable, values):
            min_val = min(values)
            return [(v - min_val) / (gv.max_value - gv.min_value) for v in values]

        unique_values_per_variables = {}
        values_per_variables = {}
        for c in population:
            if c.is_immigrant:
                continue
            for g in c:
                unique_var_values = unique_values_per_variables.get(
                    g.genetic_variable, None
                )
                var_values = values_per_variables.get(g.genetic_variable, None)
                if unique_var_values is None:
                    unique_var_values = set()
                    unique_values_per_variables[g.genetic_variable] = unique_var_values
                if var_values is None:
                    var_values = []
                    values_per_variables[g.genetic_variable] = var_values
                unique_var_values.add(g.variable_value)
                var_values.append(g.variable_value)
        for key in unique_values_per_variables:
            if len(unique_values_per_variables[key]) == 1:
                key.stacked = True
            else:
                key.stacked = False
        for key in values_per_variables:
            if key.stacked:
                key.relative_standard_deviation = 0.0
                continue
            scaled_values = scale_values(key, values_per_variables[key])
            range = max(scaled_values) - min(scaled_values)
            if range == 0:
                key.relative_standard_deviation = 0.0
            else:
                rel_st_dev = stat.stdev(scaled_values) / ga_utils.average(scaled_values)
                key.relative_standard_deviation = range / rel_st_dev



from gadapt.ga_model.chromosome import Chromosome
from gadapt.ga_model.gene import Gene
import random

from gadapt.operations.mutation.chromosome_mutation.base_chromosome_mutator import (
    BaseChromosomeMutator,
)


class RandomChromosomeMutator(BaseChromosomeMutator):

    """
    Random mutation of the chromosome.
    """

    def _mutate_chromosome(self, c: Chromosome, number_of_mutation_genes: int):
        if number_of_mutation_genes == 0:
            return
        genes_to_mutate = list(c)
        random.shuffle(genes_to_mutate)
        var_num = random.randint(1, number_of_mutation_genes)
        for g in genes_to_mutate[:var_num]:
            self._set_random_value(g, c)
        return var_num

    def _set_random_value(self, g: Gene, c: Chromosome):
        g.variable_value = round(
            g.genetic_variable.make_random_value(), g.genetic_variable.decimal_places
        )
        self._gene_mutated(g, c)



from abc import ABC, abstractmethod
from gadapt.ga_model.gene import Gene


class BaseGeneCombination(ABC):

    """
    Base class for gene combination
    """

    def combine(self, mother_gene: Gene, father_gene: Gene):
        """
        Combines two genes and returns two offsprings.
        Args:
            mother_gene (Gene): First gene to combine
            father_gene (Gene): Second gene to combine
        """
        return self._combine_genes(mother_gene, father_gene)

    @abstractmethod
    def _combine_genes(self, mother_gene: Gene, father_gene: Gene):
        pass



from abc import ABC, abstractmethod
from typing import List, Tuple
from gadapt.ga_model.chromosome import Chromosome


class BaseParentSelector(ABC):
    """
    Base Parent Selector

    Selects individuals for mating from the population
    """

    def select_mates(self, population) -> List[Tuple[Chromosome, Chromosome]]:
        """
        Selects and returns individuals for the crossover from the population
        Args:
            population: the population for the mates selection
        """
        return self._select_mates_from_population(population)

    @abstractmethod
    def _select_mates_from_population(
        self, population
    ) -> List[Tuple[Chromosome, Chromosome]]:
        pass



from abc import ABC, abstractmethod


class BasePopulationImmigrator(ABC):

    """
    Base class for population immigration
    """

    def immigrate(self, population):
        """
        Immigrates chromosomes into the population
        Args:
            population: Population to immigrate new chromosomes
        """
        self._immigrate_population(population)

    @abstractmethod
    def _immigrate_population(self, population):
        pass



from gadapt.operations.exit_check.base_exit_checker import BaseExitChecker


class MinCostExitChecker(BaseExitChecker):

    """
    Exit check based on minimal cost.
    The GA exits when there is no improvement in the minimal cost in a
    defined number of iterations.
    """

    def _is_exit(self, population):
        if population is None:
            raise Exception("population must not be null")
        return population.min_cost >= population.previous_min_cost



from gadapt.factory.ga_base_factory import BaseGAFactory
from gadapt.operations.exit_check.avg_cost_exit_checker import AvgCostExitChecker
from gadapt.operations.exit_check.base_exit_checker import BaseExitChecker
from gadapt.operations.exit_check.min_cost_exit_checker import MinCostExitChecker
from gadapt.operations.exit_check.requested_cost_exit_checker import RequestedCostExitChecker
from gadapt.operations.cost_finding.base_cost_finder import BaseCostFinder
from gadapt.operations.cost_finding.elitism_cost_finder import ElitismCostFinder
from gadapt.operations.crossover.base_crossover import BaseCrossover
from gadapt.operations.crossover.uniform_crossover import UniformCrossover
from gadapt.ga_model.ga_options import GAOptions
from gadapt.operations.immigration.chromosome_immigration.base_chromosome_immigrator import (
    BaseChromosomeImmigrator,
)
from gadapt.operations.immigration.chromosome_immigration.random_chromosome_immigrator import (
    RandomChromosomeImmigrator,
)
from gadapt.operations.immigration.population_immigration.base_population_immigrator import (
    BasePopulationImmigrator,
)
from gadapt.operations.immigration.population_immigration.common_population_immigrator import (
    CommonPopulationImmigrator,
)
from gadapt.operations.mutation.chromosome_mutation.base_chromosome_mutator import (
    BaseChromosomeMutator,
)
from gadapt.operations.mutation.chromosome_mutation.cross_diversity_chromosome_mutator import (
    CrossDiversityChromosomeMutator,
)
from gadapt.operations.mutation.chromosome_mutation.random_chromosome_mutator import (
    RandomChromosomeMutator,
)
from gadapt.operations.mutation.population_mutation.base_population_mutator import (
    BasePopulationMutator,
)
from gadapt.operations.mutation.population_mutation.composed_population_mutator import (
    ComposedPopulationMutator,
)
from gadapt.operations.mutation.population_mutation.cost_diversity_population_mutator import (
    CostDiversityPopulationMutator,
)
from gadapt.operations.mutation.population_mutation.parent_diversity_population_mutator import (
    ParentDiversityPopulationMutator,
)
from gadapt.operations.mutation.population_mutation.random_population_mutator import (
    RandomPopulationMutator,
)
from gadapt.operations.parent_selection.base_parent_selector import BaseParentSelector
from gadapt.operations.parent_selection.sampling_parent_selector import SamplingParentSelector
from gadapt.operations.sampling.base_sampling import BaseSampling
from gadapt.operations.sampling.from_top_to_bottom_sampling import FromTopToBottomSampling
from gadapt.operations.sampling.random_sampling import RandomSampling
from gadapt.operations.sampling.roulette_wheel_sampling import RouletteWheelSampling
from gadapt.operations.sampling.tournament_sampling import TournamentSampling
from gadapt.operations.gene_combination.base_gene_combination import BaseGeneCombination
from gadapt.operations.gene_combination.blending_gene_combination import BlendingGeneCombination
from gadapt.operations.variable_update.common_variable_updater import CommonVariableUpdater
import gadapt.ga_model.definitions as definitions


class GAFactory(BaseGAFactory):
    """
    Factory implementatiopn for creating  class instances based on GA options
    """    
    
    def _get_cost_finder(self) -> BaseCostFinder:
        """
        Cost Finder instance
        """
        return ElitismCostFinder()

    def _get_population_immigrator(self) -> BasePopulationImmigrator:
        """
        Population Immigrator Instance
        """
        return CommonPopulationImmigrator()

    def _get_chromosome_immigrator(self) -> BaseChromosomeImmigrator:
        """
        Chromosome Immigrator Instance
        """
        return RandomChromosomeImmigrator()

    def _get_chromosome_mutator(self) -> BaseChromosomeMutator:
        """
        Chromosome Mutator Instance
        """
        if self._ga.chromosome_mutation.strip() == definitions.CROSS_DIVERSITY:
            return CrossDiversityChromosomeMutator(
                self._get_sampling_method(
                    self._ga.cross_diversity_mutation_gene_selection
                )
            )
        elif self._ga.chromosome_mutation.strip() == definitions.RANDOM:
            return RandomChromosomeMutator()
        else:
            raise Exception("unknown chromosome mutation")

    def _population_mutator_options_validation(self):
        """
        Validates population mutator options
        """
        mutator_strings = self._ga.population_mutation.split(
            definitions.PARAM_SEPARATOR
        )
        for s in mutator_strings:
            if s.strip() not in definitions.POPULATION_MUTATOR_STRINGS:
                raise Exception(s + " is not defined as option for population mutation")

    def _make_population_mutator(
        self, population_mutator_string=None
    ) -> BasePopulationMutator:
        """
        Population Mutator Instance
        """
        self._population_mutator_options_validation()
        if population_mutator_string is None:
            population_mutator_string = self._ga.population_mutation.strip()
        if population_mutator_string.find(definitions.PARAM_SEPARATOR) > -1:
            return self._get_population_mutator_combined()
        elif population_mutator_string == definitions.COST_DIVERSITY:
            return CostDiversityPopulationMutator(
                ParentDiversityPopulationMutator(
                    self._get_sampling_method(
                        self._ga.parent_diversity_mutation_chromosome_selection
                    ),
                ),
            )
        elif population_mutator_string == definitions.PARENT_DIVERSITY:
            return ParentDiversityPopulationMutator(
                self._get_sampling_method(
                    self._ga.parent_diversity_mutation_chromosome_selection
                ),
            )
        elif population_mutator_string == definitions.RANDOM:
            return RandomPopulationMutator()
        else:
            raise Exception("unknown population mutation")
        
    def _get_population_mutator(self) -> BasePopulationMutator:
        """
        Population Mutator Instance
        """
        return self._make_population_mutator()
       

    def _get_population_mutator_combined(self) -> BasePopulationMutator:
        """
        Population Mutator Instance - combined
        """
        mutator_strings = [
            ms.strip()
            for ms in self._ga.population_mutation.split(definitions.PARAM_SEPARATOR)
        ]
        if self._is_cost_diversity_random(mutator_strings):
            return CostDiversityPopulationMutator(
                self._options, RandomPopulationMutator(self._options)
            )
        if self._is_cost_diversity_parent_diversity(mutator_strings):
            return CostDiversityPopulationMutator(
                ParentDiversityPopulationMutator(
                    self._get_sampling_method(
                        self._ga.parent_diversity_mutation_chromosome_selection
                    ),
                ),
            )
        if self._is_cost_diversity_parent_diversity_random(mutator_strings):
            composedPopulationMutator = ComposedPopulationMutator(self._options)
            composedPopulationMutator.append(
                ParentDiversityPopulationMutator(
                    self._get_sampling_method(
                        self._ga.parent_diversity_mutation_chromosome_selection
                    ),
                )
            )
            composedPopulationMutator.append(RandomPopulationMutator(self._options))
            return CostDiversityPopulationMutator(
                composedPopulationMutator
            )
        composedPopulationMutator = ComposedPopulationMutator(self._options)
        for ms in mutator_strings:
            composedPopulationMutator.append(self._make_population_mutator(ms))
        return composedPopulationMutator

    def _is_cost_diversity_random(self, mutator_strings: list):
        """
        Is population mutator cost diversity and random
        """
        if (
            len(mutator_strings) == 2
            and definitions.COST_DIVERSITY in mutator_strings
            and definitions.RANDOM in mutator_strings
        ):
            return True
        return False

    def _is_cost_diversity_parent_diversity(self, mutator_strings: list):
        """
        Is population mutator cost diversity and parent diversity
        """
        if (
            len(mutator_strings) == 2
            and definitions.COST_DIVERSITY in mutator_strings
            and definitions.PARENT_DIVERSITY in mutator_strings
        ):
            return True
        return False

    def _is_cost_diversity_parent_diversity_random(self, mutator_strings: list):
        """
        Is population mutator cost diversity, parent diversity and random
        """
        if (
            len(mutator_strings) == 3
            and definitions.COST_DIVERSITY in mutator_strings
            and definitions.PARENT_DIVERSITY in mutator_strings
            and definitions.RANDOM in mutator_strings
        ):
            return True
        return False

    def _get_parent_selector(self) -> BaseParentSelector:
        """
        Parent Selector Instance
        """
        return SamplingParentSelector(
            self._get_sampling_method(self._ga.parent_selection)
        )

    def _get_sampling_method(self, str) -> BaseSampling:
        """
        Sampling Methos Instance
        """
        str_value = str
        sampling_method_strings = str.split(definitions.PARAM_SEPARATOR)
        other_value = None
        if len(sampling_method_strings) > 1:
            str_value = sampling_method_strings[0]
            try:
                other_value = int(sampling_method_strings[1])
            except Exception:
                pass
        if str_value == definitions.TOURNAMENT:
            return TournamentSampling(other_value)
        elif str_value == definitions.FROM_TOP_TO_BOTTOM:
            return FromTopToBottomSampling()
        elif str_value == definitions.RANDOM:
            return RandomSampling()
        return RouletteWheelSampling()

    def _get_gene_combination(self) -> BaseGeneCombination:
        """
        Gene Combination Instance
        """
        return BlendingGeneCombination()

    def _get_exit_checker(self) -> BaseExitChecker:
        """
        Exit Checker Instance
        """
        if self._ga.exit_check == definitions.AVG_COST:
            return AvgCostExitChecker(self._ga.max_attempt_no)
        if self._ga.exit_check == definitions.MIN_COST:
            return MinCostExitChecker(self._ga.max_attempt_no)
        return RequestedCostExitChecker(self._ga.requested_cost)

    def _get_crossover(self) -> BaseCrossover:
        """
        Crossover Instance
        """
        return UniformCrossover(self.get_gene_combination(), self.get_chromosome_mutator(), self.get_chromosome_immigrator())

    def _get_variable_updater(self):
        """
        Variable Updater Instance
        """
        return CommonVariableUpdater()



from typing import List, Tuple
from gadapt.ga_model.chromosome import Chromosome
from gadapt.operations.parent_selection.base_parent_selector import BaseParentSelector
from gadapt.operations.sampling.base_sampling import BaseSampling


class SamplingParentSelector(BaseParentSelector):

    """
    Parent Selector based on sampling.
    Sampling is the algorithm for extracting a sample from the population,
    based on specific value of the chromosme.
    In this case sampling depends on cost value.

    Selects mates for mating from the population
    """

    def __init__(self, sampling: BaseSampling) -> None:
        super().__init__()
        self._sampling = sampling

    def _select_mates_from_population(
        self, population
    ) -> List[Tuple[Chromosome, Chromosome]]:
        working_chromosomes = self._sampling.get_sample(
            population.chromosomes, len(population), lambda c: c.cost_value
        )
        list_of_mates: List[Tuple[Chromosome, Chromosome]] = []
        while len(working_chromosomes) > 1:
            c1 = working_chromosomes.pop(0)
            c2 = working_chromosomes.pop(0)
            list_of_mates.append((c1, c2))
        return list_of_mates



from typing import Tuple
from gadapt.ga_model.chromosome import Chromosome
from gadapt.operations.crossover.base_crossover import BaseCrossover
from gadapt.ga_model.gene import Gene
from gadapt.operations.immigration.chromosome_immigration.base_chromosome_immigrator import (
    BaseChromosomeImmigrator,
)
from gadapt.operations.mutation.chromosome_mutation.base_chromosome_mutator import (
    BaseChromosomeMutator,
)
from gadapt.operations.gene_combination.base_gene_combination import BaseGeneCombination


class UniformCrossover(BaseCrossover):

    """
    Uniform Crossover.
    Genes from parents' chromosomes are combined in a uniform way
    """

    def __init__(
        self,
        var_combination: BaseGeneCombination,
        mutator: BaseChromosomeMutator,
        immigrator: BaseChromosomeImmigrator,
    ):
        super(UniformCrossover, self).__init__(var_combination, mutator, immigrator)

    def _get_mother_father_genes(
        self, mother: Chromosome, father: Chromosome
    ) -> Tuple[Gene, Gene]:
        father_gene = father[self._current_gene_number]
        mother_gene = mother[self._current_gene_number]
        return mother_gene, father_gene

    def _combine(self, mother_gene: Gene, father_gene: Gene):
        if self._gene_combination is None:
            raise Exception("gene_combination must not be null!")
        return self._gene_combination.combine(mother_gene, father_gene)



from gadapt.ga_model.ga_options import GAOptions
from gadapt.operations.mutation.population_mutation.base_population_mutator import (
    BasePopulationMutator,
)


class RandomPopulationMutator(BasePopulationMutator):

    """
    Random population mutator
    """

    def __init__(self) -> None:
        super().__init__()

    def _mutate_population(self, population, number_of_mutation_chromosomes):
        if population is None:
            raise Exception("population must not be None")
        number_of_mutation_genes = population.options.number_of_mutation_genes
        unallocated_chromosomes = self._get_unallocated_chromosomes(
            population, self._sort_key_random
        )
        chromosomes_for_mutation = unallocated_chromosomes[
            :number_of_mutation_chromosomes
        ]
        for c in chromosomes_for_mutation:
            c.mutate(number_of_mutation_genes)
        return number_of_mutation_chromosomes



"""
Population
"""
from typing import List, Tuple
from gadapt.operations.exit_check.base_exit_checker import BaseExitChecker
from gadapt.operations.cost_finding.base_cost_finder import BaseCostFinder
from gadapt.operations.crossover.base_crossover import BaseCrossover
from gadapt.ga_model.chromosome import Chromosome
from gadapt.ga_model.ga_options import GAOptions
from gadapt.ga_model.gene import Gene
from gadapt.operations.immigration.chromosome_immigration.base_chromosome_immigrator import (
    BaseChromosomeImmigrator,
)
from gadapt.operations.immigration.population_immigration.base_population_immigrator import (
    BasePopulationImmigrator,
)
from gadapt.operations.mutation.chromosome_mutation.base_chromosome_mutator import (
    BaseChromosomeMutator,
)
from gadapt.operations.mutation.population_mutation.base_population_mutator import (
    BasePopulationMutator,
)
from gadapt.operations.parent_selection.base_parent_selector import BaseParentSelector
from gadapt.operations.variable_update.base_variable_updater import BaseVariableUpdater
import gadapt.adapters.string_operation.ga_strings as ga_strings
from datetime import datetime
import gadapt.ga_model.definitions as definitions


class Population:
    def __init__(
        self,
        options: GAOptions,
        chromosome_mutator: BaseChromosomeMutator,
        population_mutator: BasePopulationMutator,
        exit_checker: BaseExitChecker,
        cost_finder: BaseCostFinder,
        population_immigrator: BasePopulationImmigrator,
        chromosome_immigrator: BaseChromosomeImmigrator,
        parent_selector: BaseParentSelector,
        crossover: BaseCrossover,
        variable_updater: BaseVariableUpdater,
    ):
        """Population for the genetic algorithm. It contains a collection of\
            chromosomes, as well as additional parameters

        Args:

            options (GAOptions): Genetic Algorithm Options
            chromosome_mutator (BaseChromosomeMutator): Chromosome Mutator Instance
            population_mutator (BasePopulationMutator): Population Mutator Instance
            exit_checker (BaseExitChecker): Exit Checker Instance
            cost_finder (BaseCostFinder): Cost Finder Instance
            population_immigrator (BasePopulationImmigrator):
            Population Immigrator Instance
            chromosome_immigrator (BaseChromosomeImmigrator):
            Chromosome Immigrator Instance
            parent_selector (BaseParentSelector): Parent Selector Instance
            crossover (BaseCrossover): Crossover Instance
            variable_updater (BaseVariableUpdater): Variable Updater Instance
        """
        if options.population_size < 4:
            raise Exception("Population size 4 must be higher than 3")
        self.options = options
        self.chromosome_mutator = chromosome_mutator
        self.population_mutator = population_mutator
        self.exit_checker = exit_checker
        self.cost_finder = cost_finder
        self.population_immigrator = population_immigrator
        self.chromosome_immigrator = chromosome_immigrator
        self.parent_selector = parent_selector
        self.crossover = crossover
        self.variable_updater = variable_updater
        self._set_init_values()
        self.last_chromosome_id = 1
        self._population_generation = 0
        self.options = options
        self.chromosomes: List[Chromosome] = []
        self.generate_initial_population()
        self.start_time = datetime.now()
        self.timeout_expired = False

    def __iter__(self):
        return PopulationIterator(self)

    def __getitem__(self, index):
        return self.chromosomes[index]

    def __next__(self):
        return next(self.chromosomes)

    def __len__(self):
        return len(self.chromosomes)

    def __str__(self):
        return self._to_string()

    def get_sorted(self, key=None, reverse: bool = False):
        """Sorted list of chromosomes
        Args:
            key: Sorted key
            reverse (bool=False): is reversed
        """
        return sorted(self.chromosomes, key=key, reverse=reverse)

    def append(self, c: Chromosome):
        self.chromosomes.append(c)

    def generate_initial_population(self):
        for i in range(self.options.population_size):
            self.add_new_chromosome()

    def _to_string(self):
        return ga_strings.population_to_string(self)

    def _set_init_values(self):
        float_init_value = definitions.FLOAT_NAN
        self.avg_cost = float_init_value
        self.previous_avg_cost = float_init_value
        self.min_cost = float_init_value
        self.previous_min_cost = float_init_value

    @property
    def options(self) -> GAOptions:
        """
        Genetic algorithm options
        """
        return self._options

    @options.setter
    def options(self, value: GAOptions):
        self._options = value

    @property
    def avg_cost(self) -> float:
        """
        Average cost of the population
        """
        return self._avg_cost

    @avg_cost.setter
    def avg_cost(self, value: float):
        self._avg_cost = value

    @property
    def previous_avg_cost(self) -> float:
        """
        Previous average cost
        """
        return self._previous_avg_cost

    @previous_avg_cost.setter
    def previous_avg_cost(self, value: float):
        self._previous_avg_cost = value

    @property
    def min_cost(self):
        """
        Minimum cost
        """
        return self._min_cost

    @min_cost.setter
    def min_cost(self, value: float):
        self._min_cost = value

    @property
    def previous_min_cost(self):
        """
        Previous minimum cost
        """
        return self._previous_min_cost

    @previous_min_cost.setter
    def previous_min_cost(self, value: float):
        self._previous_min_cost = value

    @property
    def best_individual(self) -> Chromosome:
        """
        Best individual chromosome
        """
        return self._best_individual

    @best_individual.setter
    def best_individual(self, value: Chromosome):
        self._best_individual = value

    @property
    def population_generation(self):
        """
        Current generation of the population
        """
        return self._population_generation

    @population_generation.setter
    def population_generation(self, value):
        self._population_generation = value

    @property
    def chromosome_mutator(self) -> BaseChromosomeMutator:
        """
        Chromosome mutator algorithm
        """
        return self._chromosome_mutator

    @chromosome_mutator.setter
    def chromosome_mutator(self, value: BaseChromosomeMutator):
        self._chromosome_mutator = value

    @property
    def population_mutator(self) -> BasePopulationMutator:
        """
        Population mutator algorithm
        """
        return self._population_mutator

    @population_mutator.setter
    def population_mutator(self, value: BasePopulationMutator):
        self._population_mutator = value

    @property
    def cost_finder(self) -> BaseCostFinder:
        """
        Cost finding algorithm
        """
        return self._cost_finder

    @cost_finder.setter
    def cost_finder(self, value: BaseCostFinder):
        self._cost_finder = value

    @property
    def parent_selector(self):
        """
        Parent selection algorithm
        """
        return self._parent_selector

    @parent_selector.setter
    def parent_selector(self, value):
        self._parent_selector = value

    @property
    def population_immigrator(self) -> BasePopulationImmigrator:
        """
        Population immigration algorithm
        """
        return self._population_immigrator

    @population_immigrator.setter
    def population_immigrator(self, value: BasePopulationImmigrator):
        self._population_immigrator = value

    @property
    def chromosome_immigrator(self) -> BaseChromosomeImmigrator:
        """
        Chromosome immigration algorithm
        """
        return self._chromosome_immigrator

    @chromosome_immigrator.setter
    def chromosome_immigrator(self, value: BaseChromosomeImmigrator):
        self._chromosome_immigrator = value

    @property
    def crossover(self) -> BaseCrossover:
        """
        Crossover algorithm
        """
        return self._crossover

    @crossover.setter
    def crossover(self, value: BaseCrossover):
        self._crossover = value

    @property
    def variable_updater(self) -> BaseVariableUpdater:
        """
        Variable update algorithm
        """
        return self._variable_updater

    @variable_updater.setter
    def variable_updater(self, value: BaseVariableUpdater):
        self._variable_updater = value

    @property
    def exit_checker(self) -> BaseExitChecker:
        """
        Exit checking algorithm
        """
        return self._exit_checker

    @exit_checker.setter
    def exit_checker(self, value: BaseExitChecker):
        self._exit_checker = value

    def exit(self) -> bool:
        """
        Check exit from the GA
        """
        self.population_generation += 1
        return self.exit_checker.check(self)

    def immigrate(self):
        """
        Immigrates new chromosomes
        """
        self.population_immigrator.immigrate(self)

    def select_mates(self) -> List[Tuple[Chromosome, Chromosome]]:
        """
        Selects mates for pairing
        """
        return self.parent_selector.select_mates(self)

    def mate(self):
        """
        Mates chromosomes
        """
        chromosome_pairs = self.select_mates()
        for chromosome1, chromosome2 in chromosome_pairs:
            offspring1, offspring2 = self.crossover.mate(
                chromosome1, chromosome2, self.population_generation
            )
            self.add_chromosomes((offspring1, offspring2))

    def mutate(self):
        """
        Mutates chromosomes in the population
        """
        self.population_mutator.mutate(self)

    def find_costs(self):
        """
        Finds costs for chromosomes
        """
        self.previous_avg_cost = self.avg_cost
        self.previous_min_cost = self.min_cost
        self.cost_finder.find_costs(self)

    def clear(self):
        """
        Clears all chromosomes
        """
        self.chromosomes.clear()

    def clear_and_add_chromosomes(self, chromosomes: List[Chromosome]):
        """
        Clears chromosomes and adds new ones
        Args:
            chromosomes (List[Chromosome]): chromosomes to add
        """
        self.chromosomes.clear()
        self.add_chromosomes(chromosomes)

    def add_chromosomes(self, chromosomes):
        """
        Adds chromosomes to population
        Args:
            chromosomes (List[Chromosome]): chromosomes to add
        """
        for c in chromosomes:
            self.add_chromosome(c)

    def add_new_chromosome(self):
        """
        Adds new chromosomes to the population
        """
        chromosome = Chromosome(
            self.chromosome_mutator,
            self.chromosome_immigrator,
            self.population_generation,
        )
        chromosome.chromosome_generation = 1
        self.add_chromosome(chromosome)

    def add_chromosome(self, chromosome):
        """
        Adds chromosome to the population
        Args:
            chromosome: chromosome to add
        """
        if len(self) >= self.options.population_size:
            return
        if chromosome.chromosome_id is None or chromosome.chromosome_id == -1:
            chromosome.chromosome_id = self.last_chromosome_id
            self.last_chromosome_id += 1
        if len(chromosome) == 0:
            for gv in self.options.genetic_variables:
                g = Gene(gv)
                chromosome.append(g)
        self.append(chromosome)

    def update_variables(self):
        """
        Updates genetic variables
        """
        self.variable_updater.update_variables(self)


class PopulationIterator:
    def __init__(self, population):
        self.population = population
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.index < len(self.population.chromosomes):
            result = self.population.chromosomes[self.index]
            self.index += 1
            return result
        else:
            raise StopIteration



from gadapt.operations.exit_check.base_exit_checker import BaseExitChecker


class AvgCostExitChecker(BaseExitChecker):

    """
    Exit check based on average cost.
    The GA exits when there is no improvement in
    the average cost of the kept population in a defined number of iterations.
    """

    def _is_exit(self, population):
        if population is None:
            raise Exception("population must not be null")
        return population.avg_cost >= population.previous_avg_cost



from abc import ABC, abstractmethod
import math
import random
from typing import List
from gadapt.ga_model.chromosome import Chromosome
from gadapt.ga_model.ga_options import GAOptions


class BasePopulationMutator(ABC):
    def mutate(self, population):
        """
        Mutates chromosomes in the population
        Args:
            population: Population to mutate
        """        
        number_of_mutated_chromosomes = (
            population.options.number_of_mutation_chromosomes
        )
        self._mutate_population(population, number_of_mutated_chromosomes)

    def __init__(self) -> None:
        """
        Base class for mutating chromosomes in population
        Args:
            options: genetic algorithm options
        """
        super().__init__()

    @abstractmethod
    def _mutate_population(self, population, number_of_mutated_chromosomes):
        pass

    def _get_unallocated_chromosomes(
        self, population, sort_key_function=None
    ) -> List[Chromosome]:
        def unallocated_chromosomes_condition(c: Chromosome) -> bool:
            return (
                math.isnan(c.cost_value)
                and (not c.is_immigrant)
                and c.population_generation == population.population_generation
                and not c.is_mutated
            )

        lst = [c for c in population if (unallocated_chromosomes_condition(c))]
        if sort_key_function is not None:
            lst.sort(key=sort_key_function)
        return lst

    def _sort_key_random(self, c: Chromosome):
        return random.random()



from abc import ABC, abstractmethod
import logging
from datetime import datetime


class BaseExitChecker(ABC):

    """
    Base class for exit check
    Args:
        max_attempt_no (int): Maximal number of attempts with no improvement,
        for the given criteria.

            After this number of attempts with no improvements, the GA exits
    """

    def __init__(self, max_attempt_no: int) -> None:
        self.max_attempt_no = max_attempt_no
        self.attempt_no = 0

    @property
    def attempt_no(self) -> int:
        return self._attempt_no

    @attempt_no.setter
    def attempt_no(self, value: int):
        self._attempt_no = value

    def check(self, population):
        time_diff = (datetime.now() - population.start_time).total_seconds()
        if time_diff >= population.options.timeout:
            population.timeout_expired = True
            return True
        if self._is_exit(population):
            self.attempt_no += 1
        else:
            self.attempt_no = 0
        if self.attempt_no >= self.max_attempt_no:
            logging.info("function exit.")
            return True
        return False

    @abstractmethod
    def _is_exit(self, population) -> bool:
        pass
